/**
 * Battle module.
 * Handles battle preparation, looping and victory/defeat.
 */

/**
 * Sample enemy data template.
 */
var enemy = {
	name: "Smiley", exp: 5,
	hp: 20, hpMax: 20,
	mp: 5, mpMax: 5,
	str: 2, def: 2, mag: 1,
	sprite: LoadSpriteset("circle.rss")
};

/**
 * Module.
 */
var Battle = function() {
	// Choice to escape.
	// Reset to false on beginning a battle.
	var escaped = false;
	var canEscape = true;
	
	var defaultHeroDrawer = function (x, y) {
		drawStats(hero, x, y);
		Util.getSpritesetImage(GetPersonSpriteset("Hero"), "west", 0).blit(x, y);
	};
	var defaultEnemyDrawer = function (x, y) {
		drawStats(enemy, x, y);
		Util.getSpritesetImage(enemy.sprite, "alive", 0).blit(x, y);
	};
	
	function resetEnemy() {
		enemy.hp = enemy.hpMax;
		enemy.mp = enemy.mpMax;
	}
	
	function drawStats(combatant, x, y) {
		font.drawText(x - 20, y - font.getHeight() * 2, combatant.hp + " / " + combatant.hpMax +" HP");
		font.drawText(x - 20, y - font.getHeight(), combatant.mp + " / " + combatant.mpMax + " MP")
	}
	
	/**
	 * The main battle loop.
	 */
	function begin(escapable) {
		var heroTurn = true;
		escaped = false;
		
		// Reset the one enemy we've got.
		resetEnemy();
		
		// canEscape defaults to true, otherwise
		// it's whatever escapable is set to.
		if (escapable === undefined) {
			canEscape = true;
		} else {
			canEscape = !!escapable;
		}
		
		// Battle loop.
		while (!heroDead() && !enemyDead() && !escaped) {
			if (heroTurn) {
				heroAction();
			} else {
				enemyAction();
			}
			drawBattle();
			FlipScreen();
			heroTurn = !heroTurn;
		}
		
		// Battle end.
		if (heroDead()) {
			handleDefeat();
		} else if (enemyDead()) {
			handleVictory();
		} else if (escaped) {
			// Do nothing.
		}
	}
	
	/**
	 * Draw the main components of the battle:
	 * - Map
	 * - Translucent box
	 * - Enemy sprite
	 * - Player sprite
	 * 
	 * heroDrawer and enemyDrawer are functions that take
	 * coordinates as to where the sprites should go.
	 * Default drawers indicate status, custom-supplied
	 * drawers typically show action poses.
	 */
	function drawBattle(heroDrawer, enemyDrawer) {
		if (IsMapEngineRunning()) {
			RenderMap();
		}
		
		Rectangle(
			0, 0,
			GetScreenWidth(), GetScreenHeight(),
			CreateColor(0, 0, 0, 128)
		);
		
		// Hero on the right, enemy on the left.
		var hd = heroDrawer ? heroDrawer : defaultHeroDrawer;
		var ed = enemyDrawer ? enemyDrawer : defaultEnemyDrawer;
		
		hd(GetScreenWidth() * 2 / 3, GetScreenHeight() / 2);
		ed(GetScreenWidth() / 3, GetScreenHeight() / 2);
	}
	
	/**
	 * Returns true if the hero is dead.
	 */
	function heroDead() {
		return hero.hp <= 0;
	}
	
	/**
	 * Returns true if the enemy is dead.
	 */
	function enemyDead() {
		return enemy.hp <= 0;
	}
	
	/**
	 * HANDLE HERO ACTION!
	 * - Player chooses an action.
	 * - Player performs action.
	 */
	function heroAction() {
		var x = GetScreenWidth() - 100;
		var y = GetScreenHeight() / 2 + 30;
		var menu = new DefaultMenu(x, y, 120, 120);
		menu.mouseSupport = false;
		menu.escapeable = false;
		
		// Battle menu rendering.
		var menuPreRender = menu.preRender;
		menu.preRender = function () {
			TextBox.drawSay("A " + enemy.name + " has appeared!\nCommand?", drawBattle);
			menuPreRender.apply(menu);
		};
		
		// Fight command
		menu.addText("Fight", function () {
			heroAttack();
			menu.done = true;
		});
		
		// Magic command
		menu.addText("Magic", function () {
			Magic.menu(x, y, drawBattle, true);
			if (Magic.magicUsed) {
				menu.done = true;
			}
		});
		
		// Item command
		menu.addText("Item", function () {
			Items.menu(x, y, drawBattle, true)
			if (Items.itemUsed) {
				menu.done = true;
			}
		});
		
		// Escape command
		if (canEscape) {
			menu.addText("Escape", function () {
				escaped = true;
				menu.done = true;
			});
		}
		
		menu.execute();
	}
	
	function heroAttack() {
		var effect = enemyEffect({normal: -hero.str});
		TextBox.say("Hero attacks! " + (-effect) + " damage inflicted.", drawBattle);
	}
	
	function heroEffect(affect) {
		var oldHp = hero.hp;
		hero.hp += affect.normal;
		if (hero.hp > hero.hpMax) hero.hp = hero.hpMax;
		if (hero.hp < 0) hero.hp = 0;
		return hero.hp - oldHp;
	}
	
	/**
	 * HANDLE ENEMY ACTION!
	 */
	function enemyAction() {
		var effect = heroEffect({normal: -enemy.str});
		TextBox.say("Enemy attacks! " + (-effect) + " damage taken.", drawBattle);		
	}
	
	function enemyEffect(affect) {
		var oldHp = enemy.hp;
		enemy.hp += affect.normal;
		if (enemy.hp > enemy.hpMax) enemy.hp = enemy.hpMax;
		if (enemy.hp < 0) enemy.hp = 0;
		return enemy.hp - oldHp;
	}
	
	/**
	 * HANDLE DEFEAT.
	 */
	function handleDefeat() {
		Rectangle(0, 0, GetScreenWidth(), GetScreenHeight(), CreateColor(0, 0, 0));
		font.drawText(
			(GetScreenWidth() - font.getStringWidth("SUDDEN END")) / 2,
			(GetScreenHeight() - font.getHeight()) / 2,
			"SUDDEN END"
		);
		FlipScreen();
		GetKey();
		ExitMapEngine();
	}
	
	/**
	 * HANDLE VICTORY.
	 */
	function handleVictory() {
		// Draw victory.
		drawBattle();
		font.drawText(
			(GetScreenWidth() - font.getStringWidth("VICTOLY")) / 2,
			(GetScreenHeight() - font.getHeight()) / 2,
			"VICTOLY"
		);
		FlipScreen();
		GetKey();
		
		// Handle experience and level up.
		addExperience(enemy.exp);
	}
	
	function addExperience(exp) {
		hero.exp += exp;
		while (hero.exp >= hero.next) {
			++hero.level;
			hero.exp -= hero.next;
			hero.next += 10;
			levelUp();
		}
	}
	
	function levelUp() {
		var bonuses = [
			function () { hero.hpMax += 10; return "HP up!"; },
			function () { hero.mpMax += 10; return "MP up!"; },
			function () { ++hero.str; return "Strength up!"; },
			function () { ++hero.def; return "Defense up!"; },
			function () { ++hero.mag; return "Magic up!"; },
			//function () { spells[0].level++; return "Earth up!"; },
			//function () { spells[1].level++; return "Fire up!"; },
			//function () { spells[2].level++; return "Wind up!"; },
			//function () { spells[3].level++; return "Water up!"; },
			function () { spells[4].level++; return "Heart up!"; }
		];
		
		var message = bonuses[(hero.level - 1) % bonuses.length]();
		TextBox.drawSay(message, drawBattle);
		FlipScreen();
		GetKey();
	}
	
	return {
		begin: begin
	};
}();

/**
 * Convenience function for map zone random battles.
 */
Battle.check = function () {
	if (Math.random() < 0.2) {
		Battle.begin();
	}
};