/**
 * ScriptBind framework
 *
 * Automatically binds map events to specially-located script functions.
 *
 * In the file scripts/maps/MAPNAME.js (as a JS object expression):
 * - Map scripts:
 *   - mapEnter()
 *   - mapLeave()
 *   - mapLeaveNorth()
 *   - mapLeaveEast()
 *   - mapLeastSouth()
 *   - mapLeaveWest()
 * - Per-person scripts are stored under a member in the map object with
 *   the person's name, and themselves should return an object with:
 *   - create()
 *   - destroy()
 *   - talk()
 *   - touch()
 *   - generator()
 *
 * See the example maps and scripts.
 * 
 * To kick it all off (WARNING: clobbers existing default map scripts):
 *   ScriptBind.start();
 *
 * To stop it for whatever reason (not compulsory):
 *   ScriptBind.stop();
 *
 * More methods:
 * - setPathPrefix(prefix)	// Change where to find map scripts.
 * - getPathPrefix()
 * - setStrictness(strict)	// Throw exceptions on missing map/person functions.
 * - isStrict()
 *
 * Publicly readable/writeable members:
 * - loadMapState(mapName)
 * - saveMapState(mapName, mapState)
 * - loadPersonState(mapName, personName)
 * - savePersonState(mapName, personName, personState)
 * - world	// used by default map/person state handlers
 *
 * Notes:
 * - ScriptBind.world by default contains the world, map and person states,
 *   so if you save and load that, you can preserve the world state between
 *   game runs.
 * - For global world variables, store them as members of
 *   ScriptBind.world.state, which is an object reserved for that purpose.
 * - scripts/maps/BlockMaze.js contains a simple script example, while
 *   scripts/maps/TestIsland.js shows state saving/loading and the API.
 *
 * TODO:
 * - Ignore persons that persist between maps.
 * - Augment rather than replace existing default map scripts.
 *
 * -- tunginobi
 */

var ScriptBind = (function () {
	/********************************
	 * FUNCTIONAL PROGRAMMING TOOLS *
	 ********************************/
	function forEach(array, func) {
		for (var i = 0; i < array.length; ++i)
			func(array[i]);
	}
	
	function filter(condition, array) {
		var result = [];
		forEach(array, function (element) {
			if (condition(element))
				result.push(element);
		});
		return result;
	}
	
	function reduce(combine, base, array) {
		forEach(array, function (element) {
			base = combine(base, element);
		});
		return base;
	}
	
	/*************
	 * INTERNALS *
	 *************/
	var pathPrefix = "../scripts/maps/";
	var mapObject = {};
	var strictMode = false;
	
	var mapEvents = [
		[SCRIPT_ON_ENTER_MAP, "ScriptBind.__loadMap(GetCurrentMap()); "
				+ "ScriptBind.__loadMapState(GetCurrentMap()); "
				+ "ScriptBind.__bindPersonEvents(); "
				+ "ScriptBind.__loadPersonsStates(GetCurrentMap()); "
				+ "ScriptBind.__recreatePersons(); "
				+ "ScriptBind.__trigger('mapEnter'); "],
		[SCRIPT_ON_LEAVE_MAP, "ScriptBind.__trigger('mapLeave'); "
				+ "ScriptBind.__savePersonsStates(GetCurrentMap()); "
				+ "ScriptBind.__saveMapState(GetCurrentMap()); "],
		[SCRIPT_ON_LEAVE_MAP_NORTH, "ScriptBind.__trigger('mapLeaveNorth'); "],
		[SCRIPT_ON_LEAVE_MAP_EAST, "ScriptBind.__trigger('mapLeaveEast'); "],
		[SCRIPT_ON_LEAVE_MAP_SOUTH, "ScriptBind.__trigger('mapLeaveSouth'); "],
		[SCRIPT_ON_LEAVE_MAP_WEST, "ScriptBind.__trigger('mapLeaveWest'); "]
	];
	
	// Loads, slurps and evaluates a file, and returns the result.
	function evalFile(path) {
		var file = OpenRawFile(path);
		var script = CreateStringFromByteArray(file.read(file.getSize()));
		file.close();
		return eval(script);
	}
	
	// Find the base of a filename, i.e no extension.
	function baseName(path) {
		function lastIndexOfOrEnd(c) {
			var index = path.indexOf(c);
			return index != -1 ? index : path.length;
		}
		
		var periodPos = lastIndexOfOrEnd('.');
		return path.slice(0, periodPos);
	}
	
	// Checks if a member belongs to an object.
	function memberIn(object, member) {
		return Object.prototype.hasOwnProperty.call(object, member)
				&& Object.prototype.propertyIsEnumerable.call(object, member);
	}
	
	// Make a new "thing" member in object if it's not already there.
	function establish(object, thing) {
		if (!memberIn(object, thing)) {
			object[thing] = {};
		}
	}
	
	// Return a true array from a pseudo-array,
	// optionally skipping some arguments.
	function argumentTail(args, skip) {
		if (!skip) skip = 0;
		var result = [];
		for (var i = skip; i < args.length; ++i)
			result.push(args[i]);
		return result;
	}
	
	// Get persons with names, but not the player.
	function getImportantPersons() {
		return filter(function (p) {
			return p != "" && IsInputAttached() && p != GetInputPerson();
		}, GetPersonList());
	}
	
	// Determines if a member of a map is a map function.
	function isMapFunction(funcName) {
		return reduce(function (base, current) {
			return base || (funcName == current);
		}, false, ["mapEnter", "mapLeave", "mapLeaveNorth",
				"mapLeaveEast", "mapLeaveSouth", "mapLeaveWest",
				"save", "load"]);
	}
	
	/*************
	 * INTERFACE *
	 *************/
	return {
		/**
		 * Starts ScriptBind automatic event binding.
		 */
		start: function () {
			forEach(mapEvents, function (e) {
				SetDefaultMapScript(e[0], e[1]);
			});
		},
		
		/**
		 * Stops ScriptBind automatic event binding.
		 */
		stop: function () {
			forEach(mapEvents, function (e) {
				SetDefaultMapScript(e[0], "");
			});
		},
		
		/**
		 * Change/obtain the path to find map scripts.
		 */
		setPathPrefix: function (pp) { pathPrefix = pp; },
		getPathPrefix: function () { return pp; },
		
		/**
		 * Change/obtain raising of exceptions for missing functions.
		 */
		setStrictness: function (s) { strictMode = s; },
		isStrict: function () { return strictMode; },
		
		/***********************
		 * ALTERABLE INTERFACE *
		 * Defaults provided   *
		 * Change/read at will *
		 ***********************/
		
		/**
		 * Default state holder. Preserves map and person state.
		 * Holds a bunch of maps, which in turn hold persons.
		 */
		world: {state: {}, maps: {}},
		
		/**
		 * Function called when a map needs to load its state.
		 * Retrieve map data here and return it.
		 */
		loadMapState: function (mapName) {
			if (memberIn(ScriptBind.world, "maps")
					&& memberIn(ScriptBind.world.maps, mapName)
					&& memberIn(ScriptBind.world.maps[mapName], "state"))
				return ScriptBind.world.maps[mapName].state;
			return null;
		},
		
		/**
		 * Function called when it's time for a map to save its state.
		 * Take the map state and store it.
		 */
		saveMapState: function (mapName, mapState) {
			if (mapState) {
				establish(ScriptBind.world, "maps");
				establish(ScriptBind.world.maps, mapName);
				ScriptBind.world.maps[mapName].state = mapState;
			}
		},
		
		/**
		 * Function called when a person needs to load its state.
		 * Retrieve person data here and return it.
		 */
		loadPersonState: function (mapName, personName) {
			if (memberIn(ScriptBind.world, "maps")
					&& memberIn(ScriptBind.world.maps, mapName)
					&& memberIn(ScriptBind.world.maps[mapName], "persons")
					&& memberIn(ScriptBind.world.maps[mapName].persons, personName))
				return ScriptBind.world.maps[mapName].persons[personName];
			return null;
		},
		
		/**
		 * Function called when it's time for a person to save its state.
		 * Take the person state and store it.
		 */
		savePersonState: function (mapName, personName, personState) {
			if (personState) {
				establish(ScriptBind.world, "maps");
				establish(ScriptBind.world.maps, mapName);
				establish(ScriptBind.world.maps[mapName], "persons");
				ScriptBind.world.maps[mapName].persons[personName] = personState;
			}
		},
		
		/******************
		 * MOAR INTERNALS *
		 ******************/
		
		// Load the script associated with a map.
		__loadMap: function (mapName) {
			mapObject = evalFile(pathPrefix + baseName(mapName) + ".js");
		},
		
		// Load map state if it's found.
		__loadMapState: function (mapName) {
			var mapState = ScriptBind.loadMapState(mapName);
			if (mapState)
				ScriptBind.__trigger("load", mapState);
		},
		
		// Save map state.
		__saveMapState: function (mapName) {
			if (memberIn(mapObject, "save")) {
				ScriptBind.saveMapState(mapName, mapObject.save());
			} else if (strictMode) {
				throw new Error("Trigger 'save' not found in "
						+ baseName(mapName) + ".js");
			}
		},
		
		// Load persons states if they're found. (lol grammar)
		__loadPersonsStates: function (mapName) {
			for (var p in mapObject) {
				if (memberIn(mapObject, p) && !isMapFunction(p)) {
					var personState = ScriptBind.loadPersonState(mapName, p);
					if (personState)
						ScriptBind.__triggerPerson(p, "load", personState);
				}
			}
		},
		
		// Save persons states if possible.
		__savePersonsStates: function (mapName) {
			for (var p in mapObject) {
				if (memberIn(mapObject, p) && !isMapFunction(p)
						&& memberIn(mapObject[p], "save"))
					ScriptBind.savePersonState(mapName, p, mapObject[p].save());
			}
		},
		
		// Run through all persons and set their events to bind to functions.
		// Empty persons are ignored.
		__bindPersonEvents: function () {
			var events = [[SCRIPT_ON_CREATE, "create"],
					[SCRIPT_ON_DESTROY, "destroy"],
					[SCRIPT_ON_ACTIVATE_TOUCH, "touch"],
					[SCRIPT_ON_ACTIVATE_TALK, "talk"],
					[SCRIPT_COMMAND_GENERATOR, "generator"]];
			
			forEach(getImportantPersons(), function (p) {
				forEach(events, function (e) {
					SetPersonScript(p, e[0],
							"ScriptBind.__triggerPerson('" + p + "', '"
									+ e[1] + "');");
				});
			});
		},
		
		// Trigger all createPERSONNAME scripts, since now they exist.
		__recreatePersons: function () {
			forEach(getImportantPersons(), function (p) {
				CallPersonScript(p, SCRIPT_ON_CREATE);
			});
		},
		
		// Call a map script when a map event occurs.
		__trigger: function (funcName) {
			if (memberIn(mapObject, funcName)) {
				mapObject[funcName].apply(null, argumentTail(arguments, 1));
			} else if (strictMode) {
				throw new Error("Trigger '" + funcName + "' not found in "
						+ baseName(GetCurrentMap()) + ".js");
			}
		},
		
		// Call a person script when a person event occurs.
		__triggerPerson: function (person, funcName) {
			if (memberIn(mapObject, person)) {
				if (memberIn(mapObject[person], funcName)) {
					mapObject[person][funcName].apply(null, argumentTail(arguments, 2));
				} else if (strictMode) {
					throw new Error("Trigger '" + funcName + "' not found for "
							+ "person '" + person + "' in "
							+ baseName(GetCurrentMap()) + ".js");
				}
			} else if (strictMode) {
				throw new Error("Person '" + person + "' not found for trigger '"
						+ funcName + "' in " + baseName(GetCurrentMap()) + ".js");
			}
		}
	};
})();

/**
 * Save the world state.
 * Outside because it doesn't work inside for some reason.
 */
ScriptBind.saveState = function () {
	var m = GetCurrentMap();
	ScriptBind.__savePersonsStates(m);
	ScriptBind.__saveMapState(m);
}